#include "connection.h"
#include "statement.h"
#include "protocol.h"
#include "resource.h"
#include "error.h"

const int c_FIELD_NULL = FIELD_NULL;

const TCHAR c_szDriverODBCVer[6]   = _T("03.51");
const TCHAR c_szODBC_INI[]         = _T("ODBC.INI");
const TCHAR c_szDBMSVersion[12]    = _T("00.00.0000 ");           /* mask for the DBMS version */
const TCHAR c_szKeyWords[]         = _T("");
const TCHAR c_szSpecialCharacter[] = _T("_");

const TCHAR c_szDriverName[]       = _T(PACKAGE_NAME);             /* driver name */
const TCHAR c_szDriverVersion[]    = _T(PACKAGE_VERSION);         /* driver version */
const TCHAR c_szDriverFileName[]   = _T(ODBCNG_FILE_NAME);        /* lib file name */

/* bookmark column data */
const TCHAR c_szBookmarkName[] = _T("bookmark");  /* name of bookmark column */

/*-------------------------------------------------------------------------------------------------
 * ODBC ScalarFunctions
 *-------------------------------------------------------------------------------------------------
 */
const ScalarFunction c_ScalarFunctions[22] = {
	/* as is (may be zero or one parameter) */
	{_T("CURRENT_TIME"),       _T("localtime"),         NULL},
	{_T("CURRENT_TIMESTAMP"),  _T("localtimestamp"),    NULL},
	{_T("CURTIME"),            _T("localtime"),         NULL},

	/* no parameters */
	{_T("NOW"),                _T("now"),               NULL},
	{_T("DATABASE"),           _T("current_database"),  NULL},

	/* one parameter */
	{_T("ASCII"),              _T("ascii"),             NULL},
	{_T("BIT_LENGTH"),         _T("bit_length"),        NULL},
	{_T("CHAR"),               _T("chr"),               NULL},
	{_T("CHAR_LENGTH"),        _T("length"),            NULL},
	{_T("CHARACTER_LENGTH"),   _T("length"),            NULL},	
	{_T("LCASE"),              _T("lower"),             NULL},
	{_T("LTRIM"),              _T("ltrim"),             NULL},
	{_T("OCTET_LENGTH"),       _T("octet_length"),      NULL},
	{_T("RTRIM"),              _T("rtrim"),             NULL},
	{_T("UCASE"),              _T("upper"),             NULL},

	/* two parameters */
	{_T("REPEAT"),             _T("repeat"),            NULL},
	{_T("IFNULL"),             _T("coalesce"),          NULL},

	/* three parameters */
	{_T("REPLACE"),            _T("replace"),           NULL},
	
	/* isvrat */
	{_T("POSITION"),           _T("position"),          NULL},

	/* REMOVE_BRACES = sizeof this array - 3 */
	{_T("USER"),               _T("current_user"),      NULL},	
	{_T("CURRENT_DATE"),       _T("current_date"),      NULL},
	{_T("CURDATE"),            _T("current_date"),      NULL}
	
/*
	{_T("DIFFERENCE"),
	{_T("INSERT"),
	{_T("LEFT"),
	{_T("LENGTH"),
	{_T("LOCATE"),
	{_T("RIGHT"),
	{_T("SOUNDEX"),
	{_T("SPACE"),
	{_T("SUBSTRING"),	

	{_T("ABS"),
	{_T("ACOS"),
	{_T("ASIN"),
	{_T("ATAN"),
	{_T("ATAN2"),
	{_T("CEILING"),
	{_T("COS"),
	{_T("COT"),
	{_T("DEGREES"),
	{_T("EXP"),
	{_T("FLOOR"),
	{_T("LOG"),
	{_T("LOG10"),
	{_T("MOD"),
	{_T("PI"),
	{_T("POWER"),
	{_T("RADIANS"),
	{_T("RAND"),
	{_T("ROUND"),
	{_T("SIGN"),
	{_T("SIN"),
	{_T("SQRT"),
	{_T("TAN"),
	{_T("TRUNCATE"),

	{_T("DAYNAME"),
	{_T("DAYOFMONTH"),
	{_T("DAYOFWEEK"),
	{_T("DAYOFYEAR"),
	{_T("EXTRACT"),
	{_T("HOUR"),
	{_T("MINUTE"),
	{_T("MONTH"),
	{_T("MONTHNAME"),
	{_T("QUARTER"),
	{_T("SECOND"),
	{_T("TIMESTAMPDIFF"),
	{_T("WEEK"),
	{_T("YEAR"),

	{_T("CONVERT"),
	{_T("CAST")

	{_T("CONCAT"),             _T("
	*/
};


/*-------------------------------------------------------------------------------------------------
 * Column names and types for SQLTables, SQLColumns, SQLStatistics, SQLGetTypeInfo, SQLProcedures
 *-------------------------------------------------------------------------------------------------
 */
const ColumnNameType c_ProceduresColumns[8] = {
	{_T("PROCEDURE_CAT"),     PTN_VARCHAR, CATALOG_MAX_LENGTH},
	{_T("PROCEDURE_SCHEM"),   PTN_VARCHAR, SCHEMA_MAX_LENGTH},
	{_T("PROCEDURE_NAME"),    PTN_VARCHAR, PROCEDURE_MAX_LENGTH},
	{_T("NUM_INPUT_PARAMS"),  PTN_VARCHAR, 0},
	{_T("NUM_OUTPUT_PARAMS"), PTN_VARCHAR, 0},
	{_T("NUM_RESULT_SETS"),   PTN_VARCHAR, 0},
	{_T("REMARKS"),           PTN_VARCHAR, 0},
	{_T("PROCEDURE_TYPE"),    PTN_INT4,    0}
};

const ColumnNameType c_StatisticsColumns[13] = {
	{_T("TABLE_CAT"),        PTN_VARCHAR, CATALOG_MAX_LENGTH},
	{_T("TABLE_SCHEM"),      PTN_VARCHAR, SCHEMA_MAX_LENGTH},
	{_T("TABLE_NAME"),       PTN_VARCHAR, TABLE_MAX_LENGTH},
	{_T("NON_UNIQUE"),       PTN_INT2,    0},
	{_T("INDEX_QUALIFIER"),  PTN_VARCHAR, 0},
	{_T("INDEX_NAME"),       PTN_VARCHAR, 0},
	{_T("TYPE"),             PTN_INT2,    0},
	{_T("ORDINAL_POSITION"), PTN_INT2,    0},
	{_T("COLUMN_NAME"),      PTN_VARCHAR, COLUMN_MAX_LENGTH},
	{_T("ASC_OR_DESC"),      PTN_CHAR,    1},
	{_T("CARDINALITY"),      PTN_INT4,    0},
	{_T("PAGES"),            PTN_INT4,    0},
	{_T("FILTER_CONDITION"), PTN_VARCHAR, 0}
};

const ColumnNameType c_GetTypeInfoColumns[19] = {
	{_T("TYPE_NAME"),          PTN_VARCHAR, 0},
	{_T("DATA_TYPE"),          PTN_INT2,    0},
	{_T("COLUMN_SIZE"),        PTN_INT4,    0},
	{_T("LITERAL_PREFIX"),     PTN_VARCHAR, 0},
	{_T("LITERAL_SUFFIX"),     PTN_VARCHAR, 0},
	{_T("CREATE_PARAMS"),      PTN_VARCHAR, 0},
	{_T("NULLABLE"),           PTN_INT2,    0},
	{_T("CASE_SENSITIVE"),     PTN_INT2,    0},
	{_T("SEARCHABLE"),         PTN_INT2,    0},
	{_T("UNSIGNED_ATTRIBUTE"), PTN_INT2,    0},
	{_T("FIXED_PREC_SCALE"),   PTN_INT2,    0},
	{_T("AUTO_UNIQUE_VALUE"),  PTN_INT2,    0},
	{_T("LOCAL_TYPE_NAME"),    PTN_VARCHAR, 0},
	{_T("MINIMUM_SCALE"),      PTN_INT2,    0},
	{_T("MAXIMUM_SCALE"),      PTN_INT2,    0},
	{_T("SQL_DATA_TYPE"),      PTN_INT2,    0},
	{_T("SQL_DATETIME_SUB"),   PTN_INT2,    0},
	{_T("NUM_PREC_RADIX"),     PTN_INT4,    0},
	{_T("INTERVAL_PRECISION"), PTN_INT2,    0},
};

const ColumnNameType c_SpecialColumnsColumns[8] = {
	{_T("SCOPE"),          PTN_INT2,    0},
	{_T("COLUMN_NAME"),    PTN_VARCHAR, COLUMN_MAX_LENGTH},
	{_T("DATA_TYPE"),      PTN_INT2,    0},
	{_T("TYPE_NAME"),      PTN_VARCHAR, 0},
	{_T("COLUMN_SIZE"),    PTN_INT4,    0},
	{_T("BUFFER_LENGTH"),  PTN_INT4,    0},
	{_T("DECIMAL_DIGITS"), PTN_INT2,    0},
	{_T("PSEUDO_COLUMN"),  PTN_INT2,    0}
};

const ColumnNameType c_TablesColumns[5] = {
	{_T("TABLE_CAT"),   PTN_VARCHAR, CATALOG_MAX_LENGTH},
	{_T("TABLE_SCHEM"), PTN_VARCHAR, SCHEMA_MAX_LENGTH},
	{_T("TABLE_NAME"),  PTN_VARCHAR, TABLE_MAX_LENGTH},
	{_T("TABLE_TYPE"),  PTN_VARCHAR, STR_SIZEOF("SYSTEM_TABLE")},
	{_T("REMARKS"),     PTN_VARCHAR, 0}
};

const ColumnNameType c_ColumnsColumns[18] = {
	{_T("TABLE_CAT"),         PTN_VARCHAR, CATALOG_MAX_LENGTH},
	{_T("TABLE_SCHEM"),       PTN_VARCHAR, SCHEMA_MAX_LENGTH},
	{_T("TABLE_NAME"),        PTN_VARCHAR, TABLE_MAX_LENGTH},
	{_T("COLUMN_NAME"),       PTN_VARCHAR, COLUMN_MAX_LENGTH},
	{_T("DATA_TYPE"),         PTN_INT2,    0},
	{_T("TYPE_NAME"),         PTN_VARCHAR, 0},
	{_T("COLUMN_SIZE"),       PTN_INT4,    0},
	{_T("BUFFER_LENGTH"),     PTN_INT4,    0},
	{_T("DECIMAL_DIGITS"),    PTN_INT2,    0},
	{_T("NUM_PREC_RADIX"),    PTN_INT2,    0},
	{_T("NULLABLE"),          PTN_INT2,    0},
	{_T("REMARKS"),           PTN_VARCHAR, 0},
	{_T("COLUMN_DEF"),        PTN_VARCHAR, 0},
	{_T("SQL_DATA_TYPE"),     PTN_INT2,    0},
	{_T("SQL_DATETIME_SUB"),  PTN_INT2,    0},
	/* ODBC v.3.0 */
	{_T("CHAR_OCTET_LENGTH"), PTN_INT4,    0},
	{_T("ORDINAL_POSITION"),  PTN_INT4,    0},
	{_T("IS_NULLABLE"),       PTN_VARCHAR, STR_SIZEOF("YES")}
};

const ColumnNameType c_PrimaryKeysColumns[6] = {
	{_T("TABLE_CAT"),         PTN_VARCHAR, CATALOG_MAX_LENGTH},
	{_T("TABLE_SCHEM"),       PTN_VARCHAR, SCHEMA_MAX_LENGTH},
	{_T("TABLE_NAME"),        PTN_VARCHAR, TABLE_MAX_LENGTH},
	{_T("COLUMN_NAME"),       PTN_VARCHAR, COLUMN_MAX_LENGTH},
	{_T("KEY_SEQ"),           PTN_INT2,    0},
	{_T("PK_NAME"),           PTN_VARCHAR, 0}
};

const ColumnNameType c_ProcedureColumnsColumns[19] = {
	{_T("PROCEDURE_CAT"),     PTN_VARCHAR, CATALOG_MAX_LENGTH},
	{_T("PROCEDURE_SCHEM"),   PTN_VARCHAR, SCHEMA_MAX_LENGTH},
	{_T("PROCEDURE_NAME"),    PTN_VARCHAR, PROCEDURE_MAX_LENGTH},
	{_T("COLUMN_NAME"),       PTN_VARCHAR, COLUMN_MAX_LENGTH},
	{_T("COLUMN_TYPE"),       PTN_INT2,    0},
	{_T("DATA_TYPE"),         PTN_INT2,    0},
	{_T("TYPE_NAME"),         PTN_VARCHAR, 0},
	{_T("COLUMN_SIZE"),       PTN_INT4,    0},
	{_T("BUFFER_LENGTH"),     PTN_INT4,    0},
	{_T("DECIMAL_DIGITS"),    PTN_INT2,    0},
	{_T("NUM_PREC_RADIX"),    PTN_INT2,    0},
	{_T("NULLABLE"),          PTN_INT2,    0},
	{_T("REMARKS"),           PTN_VARCHAR, 0},
	{_T("COLUMN_DEF"),        PTN_VARCHAR, 0},
	{_T("SQL_DATA_TYPE"),     PTN_INT2,    0},
	{_T("SQL_DATETIME_SUB"),  PTN_INT2,    0},
	{_T("CHAR_OCTET_LENGTH"), PTN_INT4,    0},
	{_T("ORDINAL_POSITION"),  PTN_INT4,    0},
	{_T("IS_NULLABLE"),       PTN_VARCHAR, STR_SIZEOF("YES")}
};

const ColumnNameType c_ForeignKeysColumns[14] = {
	{_T("PKTABLE_CAT"),       PTN_VARCHAR, CATALOG_MAX_LENGTH},
	{_T("PKTABLE_SCHEM"),     PTN_VARCHAR, SCHEMA_MAX_LENGTH},
	{_T("PKTABLE_NAME"),      PTN_VARCHAR, TABLE_MAX_LENGTH},
	{_T("PKCOLUMN_NAME"),     PTN_VARCHAR, COLUMN_MAX_LENGTH},
	{_T("FKTABLE_CAT"),       PTN_VARCHAR, CATALOG_MAX_LENGTH},
	{_T("FKTABLE_SCHEM"),     PTN_VARCHAR, SCHEMA_MAX_LENGTH},
	{_T("FKTABLE_NAME"),      PTN_VARCHAR, TABLE_MAX_LENGTH},
	{_T("FKCOLUMN_NAME"),     PTN_VARCHAR, COLUMN_MAX_LENGTH},
	{_T("KEY_SEQ"),           PTN_INT2,    0},
	{_T("UPDATE_RULE"),       PTN_INT2,    0},
	{_T("DELETE_RULE"),       PTN_INT2,    0},
	{_T("FK_NAME"),           PTN_VARCHAR, 0},
	{_T("PK_NAME"),           PTN_VARCHAR, 0},
	{_T("DEFERRABILITY"),     PTN_INT2,    0}
};

const ColumnNameType c_TablePrivilegesColumns[7] = {
	{_T("TABLE_CAT"),         PTN_VARCHAR, CATALOG_MAX_LENGTH},
	{_T("TABLE_SCHEM"),       PTN_VARCHAR, SCHEMA_MAX_LENGTH},
	{_T("TABLE_NAME"),        PTN_VARCHAR, TABLE_MAX_LENGTH},
	{_T("GRANTOR"),           PTN_VARCHAR, USER_MAX_LENGTH},
	{_T("GRANTEE"),           PTN_VARCHAR, USER_MAX_LENGTH},
	{_T("PRIVILEGE"),         PTN_VARCHAR, STR_SIZEOF("REFERENCES")},
	{_T("IS_GRANTABLE"),      PTN_VARCHAR, STR_SIZEOF("YES")}
};

const ColumnNameType c_ColumnPrivilegesColumns[8] = {
	{_T("TABLE_CAT"),         PTN_VARCHAR, CATALOG_MAX_LENGTH},
	{_T("TABLE_SCHEM"),       PTN_VARCHAR, SCHEMA_MAX_LENGTH},
	{_T("TABLE_NAME"),        PTN_VARCHAR, TABLE_MAX_LENGTH},
	{_T("COLUMN_NAME"),       PTN_VARCHAR, COLUMN_MAX_LENGTH},
	{_T("GRANTOR"),           PTN_VARCHAR, USER_MAX_LENGTH},
	{_T("GRANTEE"),           PTN_VARCHAR, USER_MAX_LENGTH},
	{_T("PRIVILEGE"),         PTN_VARCHAR, STR_SIZEOF("REFERENCES")},
	{_T("IS_GRANTABLE"),      PTN_VARCHAR, STR_SIZEOF("YES")}
};

/*-------------------------------------------------------------------------------------------------
 * PostgreSQL data types
 *
 * WARNING! Do not change the order of this strings
 * Types are ordered by closeness to the SQL type identifier's definition
 *-------------------------------------------------------------------------------------------------
 */
const PostgreSQLDataType c_PostgreSQLDataTypes[PSQL_DATATYPES_NUMBER] = {
	/* exect numeric data types */
	{_T("smallint"),          _T("int2"),        SQL_SMALLINT,         21, 0, SQL_NULLABLE, 2, -1, -1,          5, NULL},
	{_T("integer"),           _T("int4"),        SQL_INTEGER,          23, 0, SQL_NULLABLE, 2, -1, -1,         10, NULL},
	{_T("bigint"),            _T("int8"),        SQL_BIGINT,           20, 0, SQL_NULLABLE, 2, -1, -1,         20, NULL},

	/* autoincrementing numeric */
	{_T("serial"),            _T("serial4"),     SQL_INTEGER,          23, 1, SQL_NO_NULLS, 1, -1, -1,         10, NULL},
	{_T("bigserial"),         _T("serial8"),     SQL_BIGINT,           20, 1, SQL_NO_NULLS, 1, -1, -1,         20, NULL},

	/* exact numeric of selectable precision */
	{_T("numeric"),           _T("numeric"),     SQL_NUMERIC,        1700, 0, SQL_NULLABLE, 2, 0, 1000,        28, _T("precision,scale")},

	/* floating-point numbers */
	{_T("real"),              _T("float4"),      SQL_REAL,            700, 0, SQL_NULLABLE, 2, -1, -1,          7, NULL},
	{_T("double precision"),  _T("float8"),	     SQL_DOUBLE,          701, 0, SQL_NULLABLE, 2, -1, -1,         15, NULL},
	
	/* character types */
	{_T("char"),              _T("char"),        SQL_CHAR,             18, 0, SQL_NULLABLE, 0, -1, -1,       1024, _T("length")},
	{_T("character"),         _T("bpchar"),      SQL_CHAR,           1042, 0, SQL_NULLABLE, 0, -1, -1,       1024, _T("length")},
	{_T("character varying"), _T("varchar"),     SQL_VARCHAR,        1043, 0, SQL_NULLABLE, 0, -1, -1,       8190, _T("length")},
	{_T("name"),              _T("name"),        SQL_VARCHAR,          19, 0, SQL_NO_NULLS, 0, -1, -1,         64, NULL},
	{_T("text"),              _T("text"),        SQL_LONGVARCHAR,      25, 0, SQL_NULLABLE, 0, -1, -1,      65535, NULL},

	/* binary types */
	{_T("bytea"),             _T("bytea"),       SQL_LONGVARBINARY,    17, 0, SQL_NULLABLE, 0, -1, -1,      65535, NULL},
	
	/* date and time with time zone */
	{_T("date"),              _T("date"),        SQL_TYPE_DATE,      1082, 0, SQL_NULLABLE, 0, -1, -1,         10, NULL},
	{_T("timetz"),            _T("timetz"),      SQL_TYPE_TIME,      1266, 0, SQL_NULLABLE, 0, -1, -1,         22, NULL},
	{_T("timestamptz"),       _T("timestamptz"), SQL_TYPE_TIMESTAMP, 1184, 0, SQL_NULLABLE, 0, -1, -1,         33, NULL},

	/* date and time types without time zone */
	{_T("time"),              _T("time"),        SQL_TYPE_TIME,      1083, 0, SQL_NULLABLE, 0,  0,  6,         19, _T("precision")},
	{_T("timestamp"),         _T("timestamp"),   SQL_TYPE_TIMESTAMP, 1114, 0, SQL_NULLABLE, 0,  0,  6,         30, _T("precision")},

	/* date and time for ODBC 2.0 */
	{_T("timestamptz"),       _T("timestamptz"), SQL_TIMESTAMP,      1184, 0, SQL_NULLABLE, 0,  0,  6,         33, NULL},
	{_T("timestamp"),         _T("timestamp"),   SQL_TIMESTAMP,      1114, 0, SQL_NULLABLE, 0,  0,  6,         30, _T("precision")},
	{_T("timetz"),            _T("timetz"),      SQL_TIME,           1266, 0, SQL_NULLABLE, 0,  0,  6,         22, NULL},
	{_T("date"),              _T("date"),        SQL_DATE,           1082, 0, SQL_NULLABLE, 0, -1, -1,         10, NULL},
	{_T("time"),              _T("time"),        SQL_TIME,           1083, 0, SQL_NULLABLE, 0,  0,  6,         19, _T("precision")},

	/* interval types */
	{_T("interval"),          _T("interval"),    SQL_INTERVAL,       1186, 0, SQL_NULLABLE, 0, -1, -1,        255, _T("precision")},

	/* boolean types */
	{_T("boolean"),           _T("bool"),        SQL_BIT,              16, 0, SQL_NULLABLE, 0, -1, -1,          1, NULL},

	/* not trully correct types */
	{_T("bytea"),             _T("bytea"),       SQL_BINARY,           17, 0, SQL_NULLABLE, 0, -1, -1,       1024, NULL},
	{_T("bytea"),             _T("bytea"),       SQL_VARBINARY,        17, 0, SQL_NULLABLE, 0, -1, -1,       1024, NULL},
	{_T("numeric"),           _T("numeric"),     SQL_DECIMAL,        1700, 0, SQL_NULLABLE, 2, 0, 1000,        28, _T("precision,scale")},
	{_T("double precision"),  _T("float8"),	     SQL_FLOAT,           701, 0, SQL_NULLABLE, 2, -1, -1,         15, NULL},

	/* PostgreSQL specific types */
	{_T("character varying"), _T("character_data"), SQL_VARCHAR,    17071, 0, SQL_NULLABLE, 0, -1, -1,      65535, NULL},
	{_T("character varying"), _T("sql_identifier"), SQL_VARCHAR,    17072, 0, SQL_NO_NULLS, 0, -1, -1,         64, NULL},

	{_T("inet"),              _T("inet"),            SQL_CHAR,        869, 0, SQL_NULLABLE, 0, -1, -1,         15, NULL},
	{_T("integer"),           _T("cardinal_number"), SQL_INTEGER,   17069, 0, SQL_NULLABLE, 2, -1, -1,         10, NULL},

	{_T("path"),              _T("path"),        SQL_UNKNOWN_TYPE,    602, 0, SQL_NULLABLE, 0, -1, -1,          0, NULL},
	{_T("box"),               _T("box"),         SQL_UNKNOWN_TYPE,    603, 0, SQL_NULLABLE, 0, -1, -1,          0, NULL},
	{_T("cidr"),              _T("cidr"),        SQL_UNKNOWN_TYPE,    650, 0, SQL_NULLABLE, 0, -1, -1,          0, NULL},
	{_T("circle"),            _T("circle"),      SQL_UNKNOWN_TYPE,    718, 0, SQL_NULLABLE, 0, -1, -1,          0, NULL},
	{_T("line"),              _T("line"),        SQL_UNKNOWN_TYPE,    628, 0, SQL_NULLABLE, 0, -1, -1,          0, NULL},
	{_T("lseg"),              _T("lseg"),        SQL_UNKNOWN_TYPE,    601, 0, SQL_NULLABLE, 0, -1, -1,          0, NULL},
	{_T("macaddr"),           _T("macaddr"),     SQL_UNKNOWN_TYPE,    829, 0, SQL_NULLABLE, 0, -1, -1,          0, NULL},
	{_T("money"),             _T("money"),       SQL_UNKNOWN_TYPE,    790, 0, SQL_NULLABLE, 0, -1, -1,          0, NULL},
	{_T("point"),             _T("point"),       SQL_UNKNOWN_TYPE,    600, 0, SQL_NULLABLE, 0, -1, -1,          0, NULL},
	{_T("polygon"),           _T("polygon"),     SQL_UNKNOWN_TYPE,    604, 0, SQL_NULLABLE, 0, -1, -1,          0, NULL},
	{_T("smgr"),              _T("smgr"),        SQL_UNKNOWN_TYPE,    210, 0, SQL_NULLABLE, 0, -1, -1,          0, NULL},

	/* UNICODE version for character types */
	{_T("char"),              _T("char"),        SQL_WCHAR,            18, 0, SQL_NULLABLE, 0, -1, -1,       1024, _T("length")},
	{_T("character"),         _T("bpchar"),      SQL_WCHAR,          1042, 0, SQL_NULLABLE, 0, -1, -1,       1024, _T("length")},
	{_T("character varying"), _T("varchar"),     SQL_WVARCHAR,       1043, 0, SQL_NULLABLE, 0, -1, -1,       8190, _T("length")},
	{_T("name"),              _T("name"),        SQL_WVARCHAR,         19, 0, SQL_NO_NULLS, 0, -1, -1,         64, NULL},
	{_T("text"),              _T("text"),        SQL_WLONGVARCHAR,     25, 0, SQL_NULLABLE, 0, -1, -1,      65535, NULL},

	{_T("character varying"), _T("character_data"), SQL_WVARCHAR,   17071, 0, SQL_NULLABLE, 0, -1, -1,      65535, NULL},
	{_T("character varying"), _T("sql_identifier"), SQL_WVARCHAR,   17072, 0, SQL_NO_NULLS, 0, -1, -1,         64, NULL},

	/* special types */
	{_T("void"),              _T("void"),        SQL_CHAR,           2278, 0, SQL_NO_NULLS, 0, -1, -1,          0, NULL},
	
	/* WARNING! This must be always the (last record-1) */
	{_T("oid"),               _T("oid"),         SQL_INTEGER,          26, 1, SQL_NO_NULLS, 2, -1, -1,          0, NULL},

	/* WARNING! This must be always the last record */
	{_T("unknown"),           _T("unknown"),     SQL_UNKNOWN_TYPE,    705, 0, SQL_NULLABLE_UNKNOWN, 0, -1, -1,          0, NULL}
};

/* List of all parameters in Data Source */
const DSParameter c_stDSParameters[DS_PARAMETERS_NUMBER] = {
	/* EDIT box parameters */
	{_T("DESCRIPTION"),     _T("DSN description"),    _T(""),          DESCRIPTION_MAX_LENGTH,    IDE_DESCRIPTION},     /* DESCRIPTION_PARAM     =  0 */
	/* can be used in Connection String */
	{_T("UID"),             _T("user ID"),            _T(""),          UID_MAX_LENGTH,            IDE_UID},             /* UID_PARAM             =  1 */	
	{_T("SERVER"),          _T("server"),             _T("localhost"), SERVER_MAX_LENGTH,         IDE_SERVER},          /* SERVER_PARAM          =  2 */
	{_T("DATABASE"),        _T("database"),           _T(""),          DATABASE_MAX_LENGTH,       IDE_DATABASE},        /* DATABASE_PARAM        =  3 */
	{_T("PWD"),             _T("password"),           _T(""),          PWD_MAX_LENGTH,            IDE_PWD},             /* PWD_PARAM             =  4 */
	{_T("SSL_CERTIFICATE"), _T("SSL certificate"),
#ifdef WIN32
	_T("postgresql\\postgresql.crt"),
#else
	_T("~/.postgresql/postgresql.crt"),
#endif /* WIN32 */
	                                                                   SSL_CERTIFICATE_MAX_LENGTH, IDE_CERTIFICATE},     /* CERTIFICATE_PARAM     =  5 */
	{_T("SSL_PRIVATE_KEY"), _T("SSL private key"),    _T(""),          SSL_PRIVATE_KEY_MAX_LENGTH, IDE_PRIVATE_KEY},     /* PRIVATE_KEY_PARAM     =  6 */
	{_T("SSL_PASSPHRASE"),  _T("SSL passphrase"),     _T(""),          SSL_PWD_MAX_LENGTH,         IDE_PASSPHRASE},      /* PASSPHRASE_PARAM      =  7 */
  /* EDIT box numeric parameters */
	{_T("PORT"),            _T("port"),               _T("5432"),      PORT_MAX_LENGTH,            IDE_PORT},            /* PORT_PARAM            =  8 */
	{_T("CONNECT_TIMEOUT"), _T("connection timeout"), _T(""),          CONNECTTIMEOUT_MAX_LENGTH,  IDE_CONNECT_TIMEOUT}, /* CONNECTTIMEOUT_PARAM  =  9 */
	{_T("ALLOWED_OBJECTS"), _T("allowed objects"),    _T(""),          ALLOWED_OBJECTS_MAX_LENGTH, IDE_ALLOWED_OBJECTS}, /* ALLOWED_OBJECTS_PARAM = 10 */
	/* CHECK BOX parameters */
	{_T("USE_BUFFERING"),   _T("buffer results?"),    _T("Y"),         USEBUFFERING_MAX_LENGTH,    IDC_USE_BUFFERING},   /* USEBUFFERING_PARAM    = 11 */
	/* COMBO BOX parameters */
#ifdef USE_SSL
	{_T("USE_SSL"),         _T("SSL mode"),           _T("IGNORE,ALLOW,PREFER,REQUIRE"), COMBOBOX_MAX_LENGTH, IDC_USE_SSL},  /* USESSL_PARAM      = 12 */
#else
	{_T("USE_SSL"),         _T("SSL mode"),           _T("IGNORE"),    COMBOBOX_MAX_LENGTH,        IDC_USE_SSL},         /* USESSL_PARAM          = 12 */
#endif /* USE_SSL */
	{_T("CLIENT_ENCODING"), _T("client encoding"),    _T("UNICODE,SQL_ASCII,EUC_CN,EUC_KR,EUC_TW,EUC_JP,JOHAB,LATIN1,LATIN2,LATIN3,LATIN4,LATIN5,\
LATIN6,LATIN7,LATIN8,LATIN9,LATIN10,ISO_8859_5,ISO_8859_6,ISO_8859_7,ISO_8859_8,WIN,WIN1250,WIN1256,WIN874,SJIS,UHC,GBK,TCVN,GB18030,ALT,KOI8,\
MULE_INTERNAL,TCVN"),                                                COMBOBOX_MAX_LENGTH, IDC_CLIENT_ENCODING}         /* CLIENTENCODING_PARAM  = 13 */
};



/*-------------------------------------------------------------------------------------------------
 * WARNING! Do not change the order of this strings
 *
 * '?' means parameter
 *-------------------------------------------------------------------------------------------------
 */
const Error c_Errors[] = {
	{_T(""), _T(""), 0},
	{_T("08001"), _T("Unknown host: ?:?."),                            1},  /* ERR_UNKNOWN_HOST                 =  1 */
	{_T("08001"), _T("Unable to connect: ?:?."),                       2},  /* ERR_CANT_CONNECT_TO_HOST         =  2 */
	{_T("01004"), _T("Too small buffer, ? was truncated."),            3},  /* ERR_TOO_SMALL_BUFFER             =  3 */
	{_T("24000"), _T("No cursor was opened."),                         4},  /* ERR_STMT_INVALID_CURSOR_STATE    =  4 */
	{_T("01S00"), _T("Unknown attribute in connection string: ?=?."),  5},  /* ERR_DBC_UNKNOWN_KEY              =  5 */
	{_T("HY012"), _T("Invalid transaction operation code."),           6},  /* ERR_INVALID_TRANS_OP_CODE        =  6 */
	{_T("HY092"), _T("Invalid handle type."),                          7},  /* ERR_INVALID_HANDLE_TYPE          =  7 */
	{_T("HY000"), _T("Cursor '?' already declared."),                  8},  /* ERR_STMT_CURSOR_DECLARED         =  8 */
	{_T("HY015"), _T("No cursor name avaible."),                       9},  /* ERR_STMT_NO_CURSOR_AVAIBLE       =  9 */
	{_T("HY015"), _T("Cursor already declared as '?'."),              10},  /* ERR_STMT_CURSOR_ALREADY_DECLARED = 10 */
	{_T("34000"), _T("Invalid cursor name."),                         11},  /* ERR_STMT_INVALID_CURSOR_NAME     = 11 */
	{_T("3C000"), _T("Cursor name already exists."),                  12},  /* ERR_STMT_CURSOR_NAME_IN_USE      = 12 */
	{_T("07009"), _T("ColumnNumber was greater than the number of columns in the result set."),
	                                                                  13},  /* ERR_STMT_COLUMN_NUMBER_TOO_BIG   = 13 */
	{_T("28000"), _T("Invalid login or password."),                   14},  /* ERR_DBC_INVALID_LOGIN_DATA       = 14 */
	{_T("01000"), _T("Statement wasn't prepare."),                    15},  /* ERR_STMT_NOT_PREPARED            = 15 */
	{_T("01000"), _T("Null buffer pointer."),                         16},  /* ERR_NULL_POINTER_BUFFER          = 16 */
	{_T("01000"), _T("Unknown attribute."),                           17},  /* ERR_UNKNOWN_ATTRIBUTE            = 17 */
	{_T("HY000"), _T("?"),                                            18},  /* ERR_SERVER_ERROR                 = 18 */
	{_T("01000"), _T("Statement '?' already declared."),              19},  /* ERR_STMT_STATEMENT_DECLARED      = 19 */
	{_T("01000"), _T("Not a SELECT, INSERT, DELETE or UPDATE query"), 20},  /* ERR_STMT_INVALID_STATE           = 20 */
	{_T("01000"), _T("In non-buffering mode it's impossible to determine the number of resulted rows."),
	                                                                  21},  /* ERR_STMT_ROW_COUNT_UNAVAIBLE     = 21 */
	{_T("HY000"), _T("This functionality is disabled when SQL_ATTR_USE_BOOKMARKS = SQL_UB_OFF. Change this it you want to use bookmark."),
	                                                                  22},  /* ERR_STMT_BOOKMARKS_UNUSED        = 22 */
	{_T("HY001"), _T("Not enough memory."),                           23},  /* ERR_NOT_ENOUGH_MEMORY            = 23 */
	{_T("HY016"), _T("Can't modify an implementation row descriptor."),
	                                                                  24},  /* ERR_DESC_CANT_MODIFY_IRD         = 24 */
	{_T("HY021"), _T("Inconsistent descriptor information."),         25},  /* ERR_DESC_INCONSISTENT_INFO       = 25 */
	{_T("HY024"), _T("Descriptor handle wasn't allocated not on the same connection as statement"),
	                                                                  26},  /* ERR_STMT_DESC_WRONG_CONNECTION   = 26 */
	{_T("HY017"), _T("Cannot set implicitly allocated descriptor."),  27},  /* ERR_STMT_DESC_WRONG_ALLOC_TYPE   = 27 */
	{_T("07002"), _T("Not all parameters were bind by application."), 28},  /* ERR_STMT_NOT_ENOUGH_PARAMETERS   = 28 */
	{_T("07009"), _T("No parameter with number ? bound."),            29},  /* ERR_STMT_PARAM_NUMBER_NOT_BINDED = 29 */
	{_T("HY000"), _T("No valid table types were specified in '?'."),  30},  /* ERR_STMT_NO_VALID_TABLE_TYPES    = 30 */
	{_T("01000"), _T("This attribute is read only!"),                 31},  /* ERR_READONLY_ATTRIBUTE           = 31 */
	{_T("HY016"), _T("Can't set specified field in this descriptor."),32},  /* ERR_DESC_INVALID_FIELD           = 32 */
	{_T("07009"), _T("Cannot modify 0 record of a parameter descriptor."),
	                                                                  33},  /* ERR_DESC_MODIFY_ZERO_REC_IN_PD   = 33 */
	{_T("07009"), _T("Cannot read 0 record of a parameter descriptor."),
	                                                                  34},  /* ERR_DESC_READ_ZERO_REC_IN_PD     = 34 */
	{_T("HY091"), _T("This field is READ-ONLY!"),                     35},  /* ERR_DESC_READ_ONLY_FIELD         = 35 */
	{_T("HY091"), _T("Unknown field identifier."),                    36},  /* ERR_DESC_UNKNOWN_FIELD           = 36 */
	{_T("HY092"), _T("You can't set SQL_DESC_UNNAMED to SQL_NAMED."), 37},  /* ERR_DESC_CANT_SET_NAMED          = 37 */
	{_T("01000"), _T("Empty query has been sent to server and server successfully processed it."),
	                                                                  38},  /* ERR_STMT_EMPTY_QUERY             = 38 */
	{_T("01000"), _T("Error occured while sending parameters to the server."), 
	                                                                  39},  /* ERR_STMT_CANT_BIND_PARAMETERS    = 39 */
	{_T("01000"), _T("'?' can't be a NULL pointer or empty string."), 40},  /* ERR_EMPTY_STRING                 = 40 */
	{_T("01000"), _T("This driver version does not allow to change current catalog"),
	                                                                  41},  /* ERR_CONN_USE_UNSUPPORTED         = 41 */
	{_T("01000"), _T("Bad query syntax."),                            42},  /* ERR_BAD_QUERY_SYNTAX             = 42 */
	{_T("01000"), _T("Cannot begin transaction."),                    43},  /* ERR_CANT_BEGIN_TRANSACTION       = 43 */
	{_T("01000"), _T("Cannot ? transaction."),                        44},  /* ERR_CANT_CLOSE_TRANSACTION       = 44 */
	{_T("01000"), _T("Cannot execute statement."),                    45},  /* ERR_STMT_CANT_EXECUTE            = 45 */
	{_T("01000"), _T("Key '?' can't be set to '?' (previously set value is '?')."),
	                                                                  46},  /* ERR_DBC_KEY_ALREADY_SET          = 46 */
	{_T("HY092"), _T("Unknown operation requested."),                 47},  /* ERR_INVALID_BULK_OPERATION       = 47 */
	{_T("01000"), _T("Unable to make insert in different tables or views."),
	                                                                  48},  /* ERR_BULK_DIFFERENT_TABLES        = 48 */
	{_T("01000"), _T("There is no columns to ?."),                    49},  /* ERR_BULK_NO_COLUMNS_TO           = 49 */
	{_T("22001"), _T("Reserved buffer was less then data length, value was right-truncated."),
	                                                                  50},  /* ERR_PUTPARAM_TRUNCATED_DATA      = 50 */
	{_T("01000"), _T("Server refused request for SSL connection."),   51},  /* ERR_SERVER_REFUSED_SSL_CONNECTION= 51 */
	{_T("01000"), _T("Error occured in SSL module"),                  52},  /* ERR_SSL_ERROR                    = 52 */
	{_T("01000"), _T("Can't get bookmark for such 'SELECT ...' query."),
	                                                                  53},  /* ERR_BOOKMARK_IN_PARSE            = 53 */
	{_T("01000"), _T("Can't prepare bookmark for query with more than one statement."),
	                                                                  54},  /* ERR_BOOKMARK_ONE_QUERY           = 54 */
	{_T("01000"), _T("Impossible to use bookmarks in non-buffering mode."),
	                                                                  55},  /* ERR_BOOKMARK_IN_BUFFERING_MODE   = 55 */
	{_T("HY107"), _T("Row value out of range"),                       56},  /* ERR_INVALID_ROWNUMBER            = 56 */
	{_T("HYC00"), _T("Driver can't support requested LockTyp."),      57},  /* ERR_UNSUPPORTED_LOCKTYPE         = 57 */
	{_T("01000"), _T("This operation is unsupported for now."),       58},  /* ERR_UNSUPPORTED_OPERATION        = 58 */
	{_T("HY000"), _T("Requested orientation is inconsistent with current cursor type."),
	                                                                  59},  /* ERR_FETCH_IMPOSSIBLE_ORIENTATION = 59 */
	{_T("HY000"), _T("Requested row is absent in resultset."),        60},  /* ERR_FETCH_NO_ROW_IN_RESULTSET    = 60 */
	{_T("HY111"), _T("SQL_ATTR_FETCH_BOOKMARK_PTR is a bad pointer."),61},  /* ERR_FETCH_INVALID_BOOKMARK_VALUE = 61 */
	{_T("01000"), _T("Backend reported error during attemption to connect: '?'."),
	                                                                  62},  /* ERR_ERROR_DURING_CONNECT         = 62 */
	{_T("01000"), _T("Can't load required Open SSL libraries."),      63},  /* ERR_CANT_INIT_OPEN_SSL           = 63 */
	{_T("HY091"), _T("Invalid descriptor field identifier."),         64},  /* ERR_INVALID_DESC_IDENTIFIER      = 64 */
	{_T("01000"), _T("Attempt to load SSL certificate failed."),      65},  /* ERR_SSL_CERT_NOT_LOADED          = 65 */
	{_T("28000"), _T("Too long ? specified."),                        66},  /* ERR_TOO_LONG_DATA                = 66 */
	{_T("S1108"), _T("Concurrency option out of range."),             67},  /* ERR_OPTION_OUT_OF_RANGE          = 67 */
	{_T("S1107"), _T("Row value out of range."),                      68},  /* ERR_ROW_OUT_OF_RANGE             = 68 */
	{_T("01000"), _T("Your AllowedObjectsString:'?' has error."),     69},  /* ERR_BAD_ALLOWED_OBJECTS_STRING   = 69 */
	{_T("01000"), _T("Permission denied for some objects in your query."),
                                                                    70},  /* ERR_NOT_ALLOWED_OBJECT           = 70 */
	{_T("HY106"), _T("Fetch type out of range."),                     71},  /* ERR_EXTENDED_FETCH_BAD_FETCHTYPE = 71 */
	{_T("IM008"), _T("Attempt to display login dialog box failed."),  72},  /* ERR_FAILED_LOGIN_DLG             = 72 */
	{_T("IM008"), _T("Not enough information for connect."),          73},  /* ERR_NEED_LOGIN_INFO              = 73 */
	{_T("01000"), _T("Parameter can not be a function call."),        74}   /* ERR_WRONG_PARAMETER              = 74 */
};

